/*
  Project: amomii Neon Tennis
  Author: Shaun Armstrong
  Date Created: January 2023
  Last Modified: June 2024
  Version: V2

  Description:
  This code turns the Neon Tennis shield into an interactive two-player game.
  Players can hit a virtual ball back and forth using buttons on the shield.
  The game utilizes a 32-LED matrix to simulate a tennis court, an OLED screen
  for displaying scores and game status, and an RGB strip for visualizing serve power.

  Dependencies:
  - Wire.h
  - Adafruit_GFX.h
  - Adafruit_SSD1306.h
  - FastLED.h

  Usage Instructions:
  Detailed instructions can be found in the Neon Tennis section at:
  http://amomii.com/pages/downloads

  Contact Information:
  For support or inquiries, please contact: contact@amomii.com

*/

//Include the necessary libraries. You need to download the Adafruit ones, but the Wire library comes with the Arduino IDE

#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include "tennis_intro.h"

#include <FastLED.h>

// Define the number of LEDs and the data pin
#define NUM_LEDS 8
#define DATA_PIN 9

// Create an array to hold the LED colors
CRGB leds[NUM_LEDS];


#define SCREEN_WIDTH 128  // OLED display width, in pixels
#define SCREEN_HEIGHT 64  // OLED display height, in pixels

// Declaration for an SSD1306 display connected to I2C (SDA, SCL pins)
#define OLED_RESET 4  // Reset pin # (or -1 if sharing Arduino reset pin)
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

//Variables for reading the buttons
const int NUM_BUTTONS = 8;
byte button[NUM_BUTTONS] = { A0, A1, A2, A3, 13, 12, 11, 10 };
byte buttonState[NUM_BUTTONS];
byte lastButtonState[NUM_BUTTONS];


//variables for the shift registers
int latchPin1 = 3;
int clockPin1 = 2;
int dataPin1 = 4;

int latchPin2 = 6;
int clockPin2 = 5;
int dataPin2 = 7;



//create variables used to identify the buttons
const byte P1A = 0;       //player 1 A
const byte P1B = 1;       //player 1 B
const byte P1C = 2;       //player 1 C
const byte P1D = 3;       //player 1 D
const byte P2A = 4;       //player 2 A
const byte P2B = 5;       //player 2 B
const byte P2C = 6;       //player 2 C
const byte P2D = 7;       //player 2 D
const byte control1 = 8;  //middle button 1
const byte control2 = 9;  //middle button 2

byte player1ButtonPress[4];  //store player 1 presses
byte player2ButtonPress[4];  //store play 2 presses

byte ballPos[8];  //record the position of the ball

/*
   The following code is a multi-dimensional array. It has 8 rows and 2 columns ([8][2]).
   This is used for returning the ball. The two numbers on each row represents the two possible buttons
   which can be used to return the ball. E.g. if the ball is moving towards player 1 across the top
   of the court, buttons [0] or buttons[1] can be used to return the ball; hence the second row {0,1}.
   See the moveDirection function to understand how to use multi-dimensional arrays.
*/
const byte Returns[8][2] = {
  { 4, 5 },
  { 0, 1 },
  { 6, 7 },
  { 2, 3 },
  { 5, 4 },
  { 1, 0 },
  { 7, 6 },
  { 3, 2 }
};

bool ledBlink = 1;

//Names for the different paths the ball can travel
byte downTheLine1;
byte crossCourt1;
byte crossCourt2;
byte downTheLine2;

//variables to store who last hit the ball, who should return, who is serving, speed/direction...
const byte player2 = 2;
const byte player1 = 1;
bool SERVE = true;
int action;
byte playerPos;
byte player;
byte returner;
int Speed = 100;
int playerSpeed[] = { 0, 100, 100 };
byte Direction;
byte ballDirection;

//used in the checkButtons function to timeout when a button isn't pressed
int delayTime = 500;
unsigned long timer;
unsigned long previousTimer;

byte buzzer = 8;

//Variables used for scoring

byte player1Games;
byte player1Sets;

byte player2Games;
byte player2Sets;

byte player1Score;
byte player2Score;
bool deuce = false;
byte games = 1;
byte points = 1;
byte netPoint;
byte missPoint;
bool GAME = false;

//Variables used for serving
byte servePos;
byte fault;
bool servePower;
bool serveSpeed;
byte server;
bool doubleFault;
byte fastReturn;
byte slowReturn;

//Variables used for sound
const byte serveSound = 0;
const byte powerServeSound = 1;
const byte slowReturnSound = 2;
const byte fastReturnSound = 3;
const byte netSound = 4;
const byte missSound = 5;
const byte gameTune = 6;
const byte setTune = 7;
const byte matchTune = 8;
const byte faultSound = 9;
const byte doubleFaultSound = NUM_BUTTONS;
byte playSound;
byte freq;

int returnSoundDelay = 50;


void setup() {

  //Set the pinMode for the shift registers
  pinMode(latchPin1, OUTPUT);
  pinMode(clockPin1, OUTPUT);
  pinMode(dataPin1, OUTPUT);

  pinMode(latchPin2, OUTPUT);
  pinMode(clockPin2, OUTPUT);
  pinMode(dataPin2, OUTPUT);

    //make sure all LEDs are off
  updateLEDs(0, 0, 0, 0);

  //set all button pins as INPUT_PULLUPs
  for (int i = 0; i < NUM_BUTTONS; i++) {
    pinMode(button[i], INPUT_PULLUP);
  }


  //Set the buzzer pin as an OUTPUT
  pinMode(buzzer, OUTPUT);

  // Add the LED strip configuration, specifying the LED type, data pin, and color order
  FastLED.addLeds<WS2812B, DATA_PIN, GRB>(leds, NUM_LEDS);

  // Set the initial brightness of the LEDs (0-255)
  FastLED.setBrightness(50);
  FastLED.clear();
  FastLED.show();

  delay(1000);

  //Initiate the OLED
  display.begin(SSD1306_SWITCHCAPVCC, 0x3C);
  delay(1000);
  display.setRotation(2);       //Rotate the screen to the correct orientation
  display.clearDisplay();       //Clear the OLED
  display.display();
  display.setTextColor(WHITE);  //Set the text color
  tennisIntro();
  updateOLED(2, 12, 20, false, "player 1");  //Call the updateOLED function
  updateOLED(2, 14, 35, false, "to serve");  //Call the updateOLED function

  //Start serial communication
  //Serial.begin(9600);
}

void tennisIntro() {

  //Variables for intro animation
  int xPos = 6;
  float yPos = 22;
  float ballSize = 40;
  int tenX = 128;
  float tenY = 0;
  int ii = 20;
  int yVelocity = 0;
  int Y = 0;
  int changeY = NUM_BUTTONS;

  for (int i = 0; i < 20; i += 2) {
    tenX = tenX - i;
    tenY = pow(tenX, 1.2) / NUM_BUTTONS;
    ballSize -= 3;
    yPos -= 1;
    display.drawBitmap(tenX, tenY, tennis_logo, 128, 64, WHITE);
    display.fillCircle(xPos, yPos, ballSize, WHITE);      //x,y,radius
    display.fillCircle(xPos, yPos, ballSize - 2, BLACK);  //x,y,radius
    display.display();
    display.clearDisplay();
  }

  for (int i = 0; i < 22; i++) {

    yPos = pow((xPos - 6), 2.5) / 200;
    yPos += NUM_BUTTONS;
    tenX = tenX - (i + 20);
    tenY = pow(tenX, 1.2) / NUM_BUTTONS;

    display.drawBitmap(tenX, tenY, tennis_logo, 128, 64, WHITE);
    display.fillCircle(xPos, yPos, ballSize, WHITE);      //x,y,radius
    display.fillCircle(xPos, yPos, ballSize - 2, BLACK);  //x,y,radius
    display.display();
    display.fillCircle(xPos, yPos, ballSize, BLACK);  //x,y,radius
    display.drawBitmap(tenX, tenY, tennis_logo, 128, 64, BLACK);
    xPos += 3;
    ballSize += 3.8;
  }

  for (int i = 0; i < 22; i++) {

    display.drawBitmap(2, 80 - (Y + changeY), amomii_logo, 128, 64, WHITE);
    display.display();
    display.drawBitmap(2, 80 - (Y + changeY), amomii_logo, 128, 64, BLACK);

    changeY = changeY + yVelocity;
    yVelocity -= 2;
    Y += 16;
  }
  display.clearDisplay();
}
void updateLEDs(int cc1, int dtl1, int dtl2, int cc2) {
  /*
     This function is used to update which LED is lit (where the ball is).
     There will only ever be one led lit at a time.
     See On the PCB Lesson 1 for more details
  */

  digitalWrite(latchPin1, LOW);
  shiftOut(dataPin1, clockPin1, LSBFIRST, cc1);
  shiftOut(dataPin1, clockPin1, LSBFIRST, dtl1);
  digitalWrite(latchPin1, HIGH);

  digitalWrite(latchPin2, LOW);
  shiftOut(dataPin2, clockPin2, LSBFIRST, dtl2);
  shiftOut(dataPin2, clockPin2, LSBFIRST, cc2);
  digitalWrite(latchPin2, HIGH);
}

void updateOLED(int SIZE, int x, int y, bool CLEAR, String WORD) {
  /*
     This new function updates what is displayed on the oled screen. It accepts 5 parameters:
     size of text, x position, y position, clear/don't clear previous screen(true/false), words to be displayed
  */

  //if CLEAR equals true, clear the previous display, otherwise, don't clear it
  if (CLEAR == true) {
    display.clearDisplay();
  }

  display.setTextSize(SIZE);  // Set the text size
  display.setCursor(x, y);    // Set the start position of the word
  display.println(WORD);      //Print the word


  display.display();  // Update the oled display. Without this, nothing will change
}

void updateOLEDscore(byte GAMES1, byte GAMES2, byte SETS1, byte SETS2) {
  /*
     This function is similar to the previous in that it updates the oled, but it is
     used to show the games and sets only. When this function is called, it doesn't clear
     the previous screen, instead, it prints under it (that's why the y position starts at 30...
     0 is the top)

     It accepts four parameters:
     player 1 games, player 1 sets, player2 games, player2 sets.

     Note that player 2 score is printed first. This is because player 2 is at the left side of
     the oled screen and player 1 as at the right (and we write left to right).
  */
  display.setTextSize(1);    // Draw 2X-scale text
  display.setCursor(0, 30);  // x,y
  display.print("Games: ");
  display.print(GAMES2);
  display.print("   vs   ");
  display.print(GAMES1);

  display.setTextSize(1);    // Draw 2X-scale text
  display.setCursor(0, 40);  // x,y
  display.print("Sets:  ");
  display.print(SETS2);
  display.print("   vs   ");
  display.print(SETS1);

  display.display();  // Show text
}

byte serveButtons() {
  /*
     This function is similar to the checkButtons function, but it doesn't have flag and a while loop
  */

  byte buttonPress;

  for (int i = 0; i < NUM_BUTTONS; i++) {
    buttonState[i] = digitalRead(button[i]);
  }

  //buttonState[2] = digitalRead(button[2]);

  buttonPress = NUM_BUTTONS;
  for (int i = 0; i < NUM_BUTTONS; i++) {
    if (buttonState[i] == 0) {
      buttonPress = i;
    }
  }

  return (buttonPress);
}

byte checkButtons(int a) {
  /*
     When this function is called, it returns a value depending on which button is pressed.
     The program will get trapped in this function until a button is pressed, or it times out.
     The length of time for the timeout is determined by the int value in the function's parameters
  */

  previousTimer = timer = millis();  //Set previousTimer and timer to the current run time
  byte buttonPress;                  // used to store which button was pressed

  bool checkB = true;  //a flag to trap the program in the while loop
  while (checkB == true) {

    //Read all of the buttons and assign the result in the buttons array
    for (int i = 0; i < NUM_BUTTONS; i++) {
      buttonState[i] = digitalRead(button[i]);
    }

    /*
       If you are receiving faulty readings from button 3 (buttons[2]) after the 'for loop',
       include the following line to read it again, and it will resolve the issue.
    */
    //buttonState[2] = digitalRead(button[2]);

    /*
       This 'for loop' runs through each item in the buttonState array and checks to see if
       any of the values is 0 (a button has been pressed).
    */
    for (int i = 0; i < NUM_BUTTONS; i++) {
      if (buttonState[i] == 0) {
        buttonPress = i;
        /*
           We first decide what to do if the current player is player1 (player2 is returning).
           If player2 is returning, we only care about player2's buttons, hence "if (buttonPress > 3)"
           If player 2 has pressed a button, the flag 'checkB' becomes false, and we can leave the while loop
        */
        if (SERVE == false && player == player1) {
          if (buttonPress > 3) {
            checkB = false;
          }
        }
        /*
           We then do the same for player2 (player1 returning)
        */
        if (SERVE == false && player == player2) {
          if (buttonPress < 4) {
            checkB = false;
          }
        }

        /*
           If SERVE is true, we return the value of the button pressed even if it is the incorrect
           button. We will deal with that in the serve function
        */
        if (SERVE == true) {
          checkB = false;
        }
      }
    }

    /*
       Before the while loop, we said that timer and previousTimer both equal millis. But now
       in the while loop, we keep updating timer to equal millis(which increases its value with time),
       but we don't change the value of previousTime. Therefore, the difference between them
       gradually increases until, eventually, the difference between them is greater than 'a'.
       'a' is the value inputted into the parameter of the function when it was called. This is
       the 'delay' time. The greater that value of 'a' the longer it will take for timer - previousTimer
       to be greater than it.

       When the if statement becomes true (it has timed out), button press = NUM_BUTTONS and checkB
       becomes false so we can leave the while loop. NUM_BUTTONS, in our case, means no button pressed.
       NUM_BUTTONS because 7 is the last button (0-7). *NUM_BUTTONS = 8*
    */
    timer = millis();
    if (timer - previousTimer > a) {
      buttonPress = NUM_BUTTONS;
      checkB = false;
    }
  }

  return (buttonPress);  //return the value of buttonPressed(0-NUM_BUTTONS);
}

void serve() {
  FastLED.clear();
  FastLED.show();

  //Set all directions to 0
  downTheLine1 = crossCourt1 = crossCourt2 = downTheLine2 = 0;

  //If there have been an even number of games, set player and server to player2 (2).
  if (games % 2 == 0) {
    player = server = player2;
    //Each point, change the serve position to crossCourt1 or crossCourt2 and set the direction of the ball
    if (points % 2 == 0) {
      playerPos = P2C;
      servePos = crossCourt1 = B00000001;
      ballDirection = 5;
    } else {
      playerPos = P2B;
      servePos = crossCourt2 = B00000001;
      ballDirection = 3;
    }
  } else {
    //If it is not player2, set player and server to player1 (1).
    player = server = player1;
    if (points % 2 == 0) {
      playerPos = P1B;
      servePos = crossCourt1 = B10000000;
      ballDirection = 2;
    } else {
      playerPos = P1C;
      servePos = crossCourt2 = B10000000;
      ballDirection = 4;
    }
  }

  /*
     We stay trapped in the while loop while serve is true
  */
  while (SERVE == true) {
    /*
       set serveCheck to whatever checkButtons returns (0-NUM_BUTTONS) and use the value in the
       parameters to determine the speed of the blinking start LED
    */
    byte serveCheck = checkButtons(400);

    /*
       If the correct button was pressed, change SERVE to false so we can leave the while loop.
        playerPos corresponds to the button it is close to.
    */
    if (serveCheck == playerPos) {
      SERVE = false;
    }
    //change ledBlink from ture to false or false to true
    ledBlink = !ledBlink;

    // if ledBlink is false, turn off all leds
    if (ledBlink == false) {
      updateLEDs(0, 0, 0, 0);
    }
    //if ledBlink is false, run the updateLEDs function which will turn on the correct LED
    if (ledBlink == true) {
      updateLEDs(crossCourt1, downTheLine1, downTheLine2, crossCourt2);
    }
  }
  //if we have left the while loop, the correct button must have been pressed, so call the superSurve function
  superServe();
}

void superServe() {
  /*
     As the name suggests, the following function moves the ball in the correct direction, It accepts
     two parameters AC (action), DIR (direction). See the main loop to see where these values come from
  */
  //assign the correct value to returner (1 or 2)
  if (player == player1) {
    returner = player2;
  } else {
    returner = player1;
  }

  servePower = true;      //used to trap in the while loop
  byte powerCounter = 0;  //used to cound how long the serve button is pressed
  Speed = 200;            //the start speed for serving
  int beep = 400;         //the frequency for the beep sound
  byte hue = 95;

  while (servePower == true) {
    //assign the returned value from the serveButtons function to servePress
    int servePress = serveButtons();
    beep += 80;  //increase the value of beep by 80 each loop... this will make the sound get higher and higher
    /*
       Read the notes in the sound function to see how works, but for now, we just need to know the
       last two parameters. Here we input the frequency of the sound we want to make and the delay time
       In our case here, the frequency is the variable beep (which increases each loop) and
       the delay is 50. This delay determines how long between led changes, or how fast
       the ball moves. If you decrease this value, the ball will move faster making serving more difficult
       Increasing it will make it easier to serve. I find 50 is challenging enough.
    */
    sound(serveSound, beep, 50);
    /*
         If, as in the case below, servePress equals playerPos, it means the server is still holding
         the serve button down
    */
    if (servePress == playerPos) {
      shift(player);   //Shift the bit values in the correct direction
      powerCounter++;  //increment by 1

      if (player == player1) {
        leds[0] = CHSV(95, 255, 255);
        if (powerCounter < NUM_LEDS && powerCounter >= 0) {
          leds[powerCounter] = CHSV(hue, 255, 255);
          FastLED.show();
        }

      } else {
        leds[7] = CHSV(95, 255, 255);
        if (powerCounter < NUM_LEDS && powerCounter >= 0) {
          leds[7 - powerCounter] = CHSV(hue, 255, 255);
          FastLED.show();
        }
      }

      hue += 20;


      Speed -= 20;  //reduce the speed delay (make the serve faster)
    }
    /*
       If servePress equals 0, it means serveButtons() returned 0
       becasue the player has released the button
    */
    if (servePress == NUM_BUTTONS) {
      servePower = false;  //leave the while loop
      /*
         The byte value has been shifted in the player direction by 1 place each loop (in the
         while loop), it now must be shifted back in the opposite direction, that is what
         the following 'for 'loop' does (returner is the opposite direction to player...
         see the shift function for more details)
      */
      for (int i = 0; i < powerCounter; i++) {
        shift(returner);
      }
      fault = 0;
    }
    /*
       If the powerCounter reaches 8, the player has held the button for too long
    */
    if (powerCounter > 7) {
      fill_solid(leds, 8, CRGB::Red);
      FastLED.show();
      fault++;                                     //Add 1 to the fault counter (2 faults to lose a point)
      servePower = false;                          //so we can leave the while loop
      sound(faultSound, 700, 500);                 //play the fault sound
      updateOLED(3, 0, 0, true, "FAULT!");         //display on the oled
      updateOLED(1, 0, 30, false, "First Serve");  //display on the oled
    }
  }

  /*
     If we leave the while loop and power counter equals 7, it means the server has released the
     button at the perfect time when the final led is lit... power serve!
  */
  if (powerCounter == 7) {
    sound(powerServeSound, 200, 150);  //play the power serve sound
    Speed = 40;                        //set the speed delay to 40 (this is a fast serve, but you can reduce the value if you think it isn't fast enough)
  }
}

void shift(int PL) {
  /*
     Depending on the value in the parameters, we shift all of the possible directions 1 place.
     This may seem strange because the ball is only moving one direction at a time, but because
     the ball can only be in one place, all other directions' bit values will be 0. Bit shifting
     a zero value left or right is still zero, so nothing with change. The only value which will
     be affected is the one that has the ball on it. For example, if we shift the following 1 place
     to the right:

      B00000000
      B00000000
      B00010000
      B00000000

      we get:

      B00000000
      B00000000
      B00001000
      B00000000
  */
  if (PL == 1) {
    downTheLine1 = downTheLine1 >> 1;
    crossCourt1 = crossCourt1 >> 1;
    downTheLine2 = downTheLine2 >> 1;
    crossCourt2 = crossCourt2 >> 1;
  }

  if (PL == 2) {
    downTheLine1 = downTheLine1 << 1;
    crossCourt1 = crossCourt1 << 1;
    downTheLine2 = downTheLine2 << 1;
    crossCourt2 = crossCourt2 << 1;
  }
}

void score(int scorer) {
  /*
     when we call the score function, the parameter is the player who should receive the point
  */

  //I fault equals 2, the returner should get the point
  if (fault == 2) {
    //Serial.println("DOUBLE FAULT!");
    updateOLED(3, 0, 0, true, "DOUBLE FAULT!");  //display on the oled
    delay(1000);                                 //delay a second so the message on the oled can be read before it's cleared
    doubleFault = true;
    scorer = returner;
    fault = 0;
  }
  points++;

  //Add a point to the correct player
  if (scorer == 1) {
    player1Score++;
  } else if (scorer == 2) {
    player2Score++;
  }

  //Create strings (character arrays) for each player
  String PLAYER1;
  String PLAYER2;

  //if deuce equals false (more about deuce later), run the switch case to determine what should be displayed

  if (deuce == false) {
    /*
       Because in the game of tennis, the scoring is not simply 1, 2, 3..., we use this switch case
       to assign the correct word/value to the player String, for example 0 = LOVE
    */
    switch (player1Score) {
      case 0:
        PLAYER1 = "LOVE";
        break;
      case 1:
        PLAYER1 = "15";
        break;
      case 2:
        PLAYER1 = "30";
        break;
      case 3:
        PLAYER1 = "40";
        break;
      case 4:
        PLAYER1 = "WIN";
        break;
    }

    switch (player2Score) {
      case 0:
        PLAYER2 = "LOVE";
        break;
      case 1:
        PLAYER2 = "15";
        break;
      case 2:
        PLAYER2 = "30";
        break;
      case 3:
        PLAYER2 = "40";
        break;
      case 4:
        PLAYER2 = "WIN";
        break;
    }

    //Serial.print("PLAYER 1: ");
    //Serial.print(PLAYER1);
    //Serial.print("  ");
    //Serial.print("PLAYER 2: ");
    //Serial.println(PLAYER2);
    updateOLED(2, 0, 0, true, PLAYER2);    //display the player 2 string
    updateOLED(2, 60, 0, false, ":");      //display a colon between the two players
    updateOLED(2, 70, 0, false, PLAYER1);  //display the player 1 string
  }

  /*
     In tennis, if the score is tied at 3 3 (40 40), it is called deuce and a player must win by
     2 clear points. If he/she goes in the lead by 1 point, it is called advantage. The following
     code deals with that
  */
  if (player1Score == 3 && player2Score == 3) {
    deuce = true;
  }

  if (deuce == true) {
    if (player1Score == player2Score) {
      //Serial.println("DEUCE");
      updateOLED(3, 0, 0, true, "DEUCE");  //if even display deuce on the oled
    } else if (player1Score > player2Score) {
      //Serial.println("Advantage PLAYER 1");
      updateOLED(3, 0, 0, true, "Adv    PLAYER1");  //if player 1 is in the lead by 1 point, display adv player 1 on the oled
    } else if (player1Score < player2Score) {
      //Serial.println("Advantage PLAYER 2");
      updateOLED(3, 0, 0, true, "Adv    PLAYER2");  //if player 2 is in the lead by 1 point, display adv player 2 on the oled
    }
  }
  /*
     The following if statements check if a player has scored more than 3 points and
     the difference between players scores is greater than 1, if so, he/she wins the game
  */
  if (player1Score > player2Score && player1Score > 3) {
    if ((player1Score - player2Score) > 1) {
      //Serial.println ("Game, PLAYER 1");
      player1Games++;                               //If a player 1 wins a game, add 1 to player1Games
      sound(gameTune, 0, 100);                      //play the game sound
      GAME = true;                                  //make the GAME bool true (used later)
      player1Score = player2Score = 0;              //set scores back to 0 for next game
      deuce = false;                                //set deuce to false for next game
      updateOLED(1, 0, 0, true, "Game, PLAYER 1");  //display the winner on the oled
      games++;                                      //add 1 to the game counter
      sound(gameTune, 0, 0);
      //Serial.print("PLAYER 1 GAMES: ");
      //Serial.print(player1Games);
      //Serial.print("  ");
      //Serial.print("PLAYER 2 GAMES: ");
      //Serial.println(player2Games);
    }
  }

  //The code below is the same as above but for player 2
  if (player2Score > player1Score && player2Score > 3) {
    if ((player2Score - player1Score) > 1) {
      //Serial.println ("Game, PLAYER 2");
      //If a player 2 wins a game, add 1 to player2Games
      player2Games++;
      sound(gameTune, 0, 100);
      GAME = true;
      player2Score = player1Score = 0;
      deuce = false;
      updateOLED(1, 0, 0, true, "Game, PLAYER 1");
      games++;
      sound(gameTune, 0, 0);
      //Serial.print("PLAYER 1 GAMES: ");
      //Serial.print(player1Games);
      //Serial.print("  ");
      //Serial.print("PLAYER 2 GAMES: ");
      //Serial.println(player2Games);
    }
  }

  /*
     To win a set in tennis, a player must reach 6 games and be 2 or more games ahead the opponent
     The following deals with that
  */
  if (player1Games > 5 && player1Games - player2Games >= 2) {
    player1Sets++;           //If this is the case, add 1 to player1Sets
    sound(setTune, 0, 100);  //play the sound for winning a set
    //Print to the monitor the sets scores
    //Serial.print("PLAYER 1 SETS: ");
    //Serial.print(player1Sets);
    //Serial.print("  ");
    //Serial.print("PLAYER 2 SETS: ");
    //Serial.println(player2Sets);

    //Reset player 1 and 2's games scores to 0 for the next set
    player1Games = player2Games = 0;
  }
  //The code below is the same as above but for player 2
  if (player2Games > 5 && player2Games - player1Games >= 2) {
    //If this is the case, add 1 to player2Sets
    player2Sets++;
    sound(setTune, 0, 100);
    //Print to the monitor the sets scores
    //Serial.print("PLAYER 1 SETS: ");
    //Serial.print(player1Sets);
    //Serial.print("  ");
    //Serial.print("PLAYER 2 SETS: ");
    //Serial.println(player2Sets);
    //Reset player 1 and 2's games scores to 0 for the next set

    player1Games = player2Games = 0;
  }
  /*
     If deuce is true, we just display the words (deuce/advantage..), otherwise, we display the score
     using the updateOLEDscore function
  */
  if (deuce == false) {
    updateOLEDscore(player1Games, player2Games, player1Sets, player2Sets);
  }

  //If a player reaches 3 sets, he/she wins the game
  if (player1Sets == 3) {
    //Serial.println ("Game, set and match, PLAYER 1!");
    //display the words GAME SET AND MACTH and the player who won
    updateOLED(2, 0, 0, true, "GAME, SET,");
    updateOLED(3, 0, 20, false, "MATCH!");                        //The parameter "false" prevents the function from deleting the previous line
    updateOLED(2, 0, 45, false, "PLAYER1");                       //The parameter "false" prevents the function from deleting the previous line
    sound(matchTune, 0, 100);                                     //Play the match sound
    player1Games = player2Games = player1Sets = player2Sets = 0;  //reset everything for the next match
  }
  //As above but for player 2
  if (player2Sets == 3) {
    //Serial.println ("Game, set and match, PLAYER 2!");
    updateOLED(2, 0, 0, true, "GAME, SET,");
    updateOLED(3, 0, 20, false, "MATCH!");
    updateOLED(2, 0, 45, false, "PLAYER2");
    sound(matchTune, 0, 100);
    player1Games = player2Games = player1Sets = player2Sets = 0;
  }
  SERVE = true;
}

void moveDirection(byte AC, byte DIR) {
  /*
     If DIR equals NUM_BUTTONS, it means the findDirection() function has returned NUM_BUTTONS, which means the ball
     has gone off the court, or has not been returned, so the last player to have hit the
     ball gets the point
  */
  if (DIR == NUM_BUTTONS) {
    missPoint = player;
  } else {
    /*
       If AC equals NUM_BUTTONS, it means the checkButtons function has returned NUM_BUTTONS, no button
       has been pressed, so we simply shift in the correct direction.
    */
    if (AC == NUM_BUTTONS) {
      shift(player);
      /*
         Now we check AC (the value returned from the checkButton function) against the
         multi-dimensional array 'Returns[][]'. Remember that in a multi-dimensional array, the first
         square bracket is for the rows, and the second one is for the columns. Returns[][] was written in
         such a way that the rows correspond to the direction of the ball. There are 8 possible directions
         (4 paths left and right) and therefore 8 rows in our multi-dimensional array. For example,
         the ball traveling from left to right across the top of the court (down the line) is direction 0. The only
         two buttons good for returning the ball are the two buttons in the top right of the court,
         buttons 4 and 5. If you look at the multi-dimensional array, you will notice that row 0
         has two values, 4 and 5. If button 4 is pressed, the ball will be returned across the same
         path (down the line), but if button 5 is pressed, the ball will be returned
         diagonally (cross court). In other words, column 0 is for returning on the same path, and
         column 1 is for changing the path.
      */
    } else if (AC == Returns[DIR][0]) {
      /*
         Now we know the correct button was pressed, we check when it was pressed.
         If it was pressed when the last led was lit, we return the ball and decrease the player's
         speed delay (make the ball move faster).
      */
      if (ballPos[DIR] == fastReturn) {
        playerSpeed[returner] -= 7;
        player = returner;
        playSound = fastReturnSound;
        freq = 1000;

        //If the button was pressed when the second to last led was lit, return the ball but don't change the speed.

      } else if (ballPos[DIR] == slowReturn) {
        player = returner;
        playSound = slowReturnSound;
        freq = 1200;

        //If the button was pressed at any other time, the player who last hit the ball gets a point
      } else {
        netPoint = player;
      }
      sound(playSound, freq, returnSoundDelay);  //play the ball hit sound

      //The statement below is the same as above with the addition of the change direction function
    } else if (AC == Returns[DIR][1]) {
      if (ballPos[DIR] == fastReturn) {
        playerSpeed[returner] -= 7;
        player = returner;
        playSound = fastReturnSound;
        freq = 600;
        changePath(Returns[DIR][1]);  //call the changePath function
      } else if (ballPos[DIR] == slowReturn) {
        player = returner;
        playSound = slowReturnSound;
        freq = 300;
        changePath(Returns[DIR][1]);  //call the changePath function
      } else {
        netPoint = player;
      }
      sound(playSound, freq, returnSoundDelay);
    } else {
      //If any other button was pressed at any other time, the play who last hit the ball gets the point
      netPoint = player;
    }
  }
}

byte findDirection() {
  /*
     This function finds which direction the ball is moving
  */

  int DIR = NUM_BUTTONS;  //NUM_BUTTONS = nothing pressed

  //clear the ballPos array (make all values 0)
  for (int i = 0; i < 8; i++) {
    ballPos[i] = 0;
  }

  /*
     If this path is greater than 0, the ball is on this path and all other paths must be 0
     because the ball an only be on one path at any given time.
  */
  if (downTheLine1 > 0) {
    /*
       If the current player is player 1, the ball is moving right. As mentioned in the
       moveDirection function comments, direction 0 is downTheLine1 moving right. So we assign
       ballPos[0] the value of downTheLine1 which could be anything between B00000001-B10000000.
       The value of every other position in the ballPos array will be 0.
    */
    if (player == player1) {
      DIR = 0;
      ballPos[DIR] = downTheLine1;
    } else {
      /*
        If the current player is player 2, direction 1 is downTheLine1 moving left. So we assign
        ballPos[1] the value of downTheLine1 which could be anything between B00000001-B10000000.
        The value of every other position in the ballPos array will be 0.
      */
      DIR = 1;
      ballPos[DIR] = downTheLine1;
    }
  }

  /*
     We do the same as in the case of downTheLine1, but for crossCourt1 which deals with
     directions 2 and 5
  */

  if (crossCourt1 > 0) {
    if (player == player1) {
      DIR = 2;
      ballPos[DIR] = crossCourt1;
    } else {
      DIR = 5;
      ballPos[DIR] = crossCourt1;
    }
  }

  /*
     We do the same as in the case of downTheLine1, but for crossCourt2 which deals with
     directions 4 and 3
  */

  if (crossCourt2 > 0) {
    if (player == player1) {
      DIR = 4;
      ballPos[DIR] = crossCourt2;
    } else {
      DIR = 3;
      ballPos[DIR] = crossCourt2;
    }
  }

  /*
     We do the same as in the case of downTheLine1, but for downTheLine2 which deals with
     directions 6 and 7
  */

  if (downTheLine2 > 0) {
    if (player == player1) {
      DIR = 6;
      ballPos[DIR] = downTheLine2;
    } else {
      DIR = 7;
      ballPos[DIR] = downTheLine2;
    }
  }

  return (DIR);  //We then return DIR which will be 0-7
}

void changePath(int path) {
  /*
     This function changes the path the ball is moving on.
  */

  //First we clear the ball from all of the paths by assigning them all as 0

  downTheLine1 = crossCourt1 = downTheLine2 = crossCourt2 = 0;

  /*
     Depending on the input parameter 'path' we give the correct path a value of either
     B10000000 (starting from left) or B00000001 (starting from left).
  */
  switch (path) {
    case 0:
      downTheLine1 = B10000000;
      break;
    case 1:
      crossCourt1 = B10000000;
      break;
    case 2:
      crossCourt2 = B10000000;
      break;
    case 3:
      downTheLine2 = B10000000;
      break;
    case 4:
      downTheLine1 = B00000001;
      break;
    case 5:
      crossCourt2 = B00000001;
      break;
    case 6:
      crossCourt1 = B00000001;
      break;
    case 7:
      downTheLine2 = B00000001;
      break;
  }
}

void sound(int SOUND, int FREQ, int DELAY) {
  /*
     This function controls the sounds in the game. It accepts 3 parameters, the sound to be played,
     the frequency of the sound to be played (the note), and how long the sound should stay on (delay).

     We use 3 Arduino functions here tone, noTone and delay. The tone function accepts 2 parameters,
     the pin connected to the buzzer, and the frequency to play. The noTone function stops the buzzer,
     and accepts one parameter, the buzzer pin. We then simply add a delay between the tone function
     and the noTone function to control how long we play the tone.
  */


  /*
     If the gameTune is the sound to be played, we play a little tune using C, E, and G.
     We could have saved the frequencies for these notes as global variables, but putting the
     actual frequencies into the tone function is more reliable.

     We do the same if the setTune or the matchTune are to be played as you can see below.
  */
  if (SOUND == gameTune) {
    tone(buzzer, 261);  //c
    delay(DELAY);
    noTone(buzzer);
    tone(buzzer, 329);  //e
    delay(DELAY);
    noTone(buzzer);
    tone(buzzer, 261);  //c
    delay(DELAY);
    noTone(buzzer);
    tone(buzzer, 392);  //g
    delay(DELAY * 4);
    noTone(buzzer);
  } else if (SOUND == setTune) {
    tone(buzzer, 392);  //g
    delay(DELAY);
    noTone(buzzer);
    tone(buzzer, 329);  //e
    delay(DELAY);
    noTone(buzzer);
    tone(buzzer, 392);  //g
    delay(DELAY);
    noTone(buzzer);
    tone(buzzer, 261);  //c
    delay(DELAY * 4);
    noTone(buzzer);

  } else if (SOUND == matchTune) {
    tone(buzzer, 261);  //c
    delay(DELAY);
    noTone(buzzer);
    tone(buzzer, 261);  //c
    delay(DELAY);
    noTone(buzzer);
    delay(DELAY);
    tone(buzzer, 261);  //c
    delay(DELAY);
    noTone(buzzer);
    tone(buzzer, 329);  //e
    delay(DELAY * 2);
    noTone(buzzer);
    tone(buzzer, 261);  //c
    delay(DELAY * 2);
    noTone(buzzer);
    tone(buzzer, 392);  //g
    delay(DELAY * 4);
    noTone(buzzer);
  } else if (SOUND == powerServeSound) {

    /*
       If the sound to be played is the powerServeSound, we use a 'for loop' to make a
       kind of powering up sound.

       We achieve this by starting the frequency in the tone function (using the variable 'beep')
       at 1000 hz and decreasing the value every 6 ms... we do this 150 times. Basically, the
       frequency of the sound starts at 1000 and quickly decreases to 400. This creates a
       pretty cool effect that sounds like power building up. It works nicely for the power serve.
    */
    int beep = 1000;
    for (int i = 150; i > 0; i--) {
      tone(buzzer, beep);
      beep -= 4;
      delay(6);
      noTone(buzzer);
    }
    //We then play a quick beep to end the power up sound
    tone(buzzer, FREQ);
    delay(DELAY);
    noTone(buzzer);
  } else {
    //In any other case we play the frequency from the parameters and delay the sound depending on the parameters
    tone(buzzer, FREQ);
    delay(DELAY);
    noTone(buzzer);
  }
}

void loop() {


  //for debugging
  bool test = false;
  while (test) {
    //Serial.println(checkButtons(500));
  }

  //If SERVE is true, we call the serve function
  if (SERVE == true) {
    serve();
    /*
       If we have left the serve function, a good serve or a fault must have happened
    */
    updateLEDs(crossCourt1, downTheLine1, downTheLine2, crossCourt2);  //Light up the correct start LED
    netPoint = 0;                                                      //set to 0 for the start of a game
    missPoint = 0;                                                     //set to 0 for the start of a game
    playerSpeed[1] = playerSpeed[2] = 100;                             //set both player speeds to 100
    /*
       The speed of the ball is determined by the serve until it has been returned, the variable
       below is for that
    */
    serveSpeed = true;
  }

  //If there was a fault, enter the if statement
  if (fault == 1) {
    SERVE = true;  //Return to the SERVE function after the loop
    //The while loop below traps the program until the serve button has been released
    while (serveButtons() == playerPos)
      ;
    return;  //Return to the top of the main loop (and back to the serve function) without running the rest of the code
  }

  //Determine who the returner is and what accounts for a fast or slow return depending on who the active player is.
  if (player == player1) {
    returner = player2;
    fastReturn = B00000001;
    slowReturn = B00000010;
  }
  if (player == player2) {
    returner = player1;
    fastReturn = B10000000;
    slowReturn = B01000000;
  }

  //Unless that ball has just been served, the speed depends on the active player
  if (serveSpeed == false) {
    Speed = playerSpeed[player];
  }

  //Assign 'action' the returned value from the checkButtons function
  action = checkButtons(Speed);

  //Assign 'Direction' the returned value from the findDirection function
  Direction = findDirection();

  //Call the moveDirection function and use 'action' and 'Direction' as its parameters
  moveDirection(action, Direction);

  //Light up the correct LED based on the value of the 4 paths (all but 1 should have the value 0)
  updateLEDs(crossCourt1, downTheLine1, downTheLine2, crossCourt2);

  //If a net point has been called, award the correct player the point and play the sound
  if (netPoint > 0) {
    //Serial.println("NET");
    score(netPoint);
    sound(netSound, 200, 150);
    sound(netSound, 150, 150);
  }

  //If a miss point has been called, award the correct player the point and play the sound
  if (missPoint > 0) {
    //Serial.println("MISS");
    score(missPoint);
    //Only play the sound if GAME is false
    if (GAME == false) {
      if (doubleFault == false) {
        sound(missSound, 400, 500);
      }
    }
    GAME = false;
    doubleFault = false;
  }

  /*
     If 'server' is the same value as 'returner', the ball has been returned so serveSpeed
     is false. The speed of the ball will now be determined by playerSpeed
  */
  if (server == returner) {
    serveSpeed = false;
  }
}
