/*
  Project: amomii Neon Tennis Test Code
  Author: Shaun Armstrong
  Date Created: July 2024
  Last Modified: July 2024
  Version: V2

  Description:
  This test code ensures all components of the Neon Tennis shield are working correctly.
  It includes tests for the OLED display, LEDs, buttons, and the amomii Glow strip.

  Dependencies:
  - Wire.h
  - Adafruit_GFX.h
  - Adafruit_SSD1306.h
  - FastLED.h

  Usage Instructions:
  Detailed instructions can be found in the Neon Tennis section at:
  http://amomii.com/pages/downloads

  Contact Information:
  For support or inquiries, please contact: contact@amomii.com

*/

#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include "images.h"
#include <FastLED.h>

// Define the number of LEDs and the data pin
#define NUM_LEDS 8
#define DATA_PIN 9

// Create an array to hold the LED colors
CRGB leds[NUM_LEDS];

#define SCREEN_WIDTH 128  // OLED display width, in pixels
#define SCREEN_HEIGHT 64  // OLED display height, in pixels

// Declaration for an SSD1306 display connected to I2C (SDA, SCL pins)
#define OLED_RESET 4  // Reset pin # (or -1 if sharing Arduino reset pin)
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

// Variables for reading the buttons
const int NUM_BUTTONS = 8;
byte button[NUM_BUTTONS] = { A0, A1, A2, A3, 13, 12, 11, 10 };
byte buttonState[NUM_BUTTONS];
byte lastButtonState[NUM_BUTTONS];

// Variables for the shift registers
int latchPin1 = 3;
int clockPin1 = 2;
int dataPin1 = 4;

int latchPin2 = 6;
int clockPin2 = 5;
int dataPin2 = 7;

byte buzzer = 8;

void setup() {
  // Set the pinMode for the shift registers
  pinMode(latchPin1, OUTPUT);
  pinMode(clockPin1, OUTPUT);
  pinMode(dataPin1, OUTPUT);

  pinMode(latchPin2, OUTPUT);
  pinMode(clockPin2, OUTPUT);
  pinMode(dataPin2, OUTPUT);

  // Make sure all LEDs are off
  updateLEDs(0, 0, 0, 0);

  // Set all button pins as INPUT_PULLUPs
  for (int i = 0; i < NUM_BUTTONS; i++) {
    pinMode(button[i], INPUT_PULLUP);
  }

  // Set the buzzer pin as an OUTPUT
  pinMode(buzzer, OUTPUT);

  // Add the LED strip configuration, specifying the LED type, data pin, and color order
  FastLED.addLeds<WS2812B, DATA_PIN, GRB>(leds, NUM_LEDS);

  // Set the initial brightness of the LEDs (0-255)
  FastLED.setBrightness(50);
  FastLED.clear();
  FastLED.show();

  delay(1000);

  // Initiate the OLED
  display.begin(SSD1306_SWITCHCAPVCC, 0x3C);
  delay(1000);
  display.setRotation(2);  // Rotate the screen to the correct orientation
  display.clearDisplay();  // Clear the OLED
  display.display();
  display.setTextColor(WHITE);  // Set the text color

  // Start serial communication
  Serial.begin(9600);
}

void updateLEDs(int cc1, int dtl1, int dtl2, int cc2) {
  /*
     This function is used to update which LED is lit (where the ball is).
     There will only ever be one led lit at a time.
     See On the PCB Lesson 1 for more details
  */

  digitalWrite(latchPin1, LOW);
  shiftOut(dataPin1, clockPin1, LSBFIRST, cc1);
  shiftOut(dataPin1, clockPin1, LSBFIRST, dtl1);
  digitalWrite(latchPin1, HIGH);

  digitalWrite(latchPin2, LOW);
  shiftOut(dataPin2, clockPin2, LSBFIRST, dtl2);
  shiftOut(dataPin2, clockPin2, LSBFIRST, cc2);
  digitalWrite(latchPin2, HIGH);
}

void tennisIntro() {
  // Variables for intro animation
  int xPos = 6;
  float yPos = 22;
  float ballSize = 40;
  int tenX = 128;
  float tenY = 0;
  int ii = 20;
  int yVelocity = 0;
  int Y = 0;
  int changeY = NUM_BUTTONS;

  for (int i = 0; i < 20; i += 2) {
    tenX = tenX - i;
    tenY = pow(tenX, 1.2) / NUM_BUTTONS;
    ballSize -= 3;
    yPos -= 1;
    display.drawBitmap(tenX, tenY, tennis_logo, 128, 64, WHITE);
    display.fillCircle(xPos, yPos, ballSize, WHITE);      // x, y, radius
    display.fillCircle(xPos, yPos, ballSize - 2, BLACK);  // x, y, radius
    display.display();
    display.clearDisplay();
  }

  for (int i = 0; i < 22; i++) {
    yPos = pow((xPos - 6), 2.5) / 200;
    yPos += NUM_BUTTONS;
    tenX = tenX - (i + 20);
    tenY = pow(tenX, 1.2) / NUM_BUTTONS;

    display.drawBitmap(tenX, tenY, tennis_logo, 128, 64, WHITE);
    display.fillCircle(xPos, yPos, ballSize, WHITE);      // x, y, radius
    display.fillCircle(xPos, yPos, ballSize - 2, BLACK);  // x, y, radius
    display.display();
    display.fillCircle(xPos, yPos, ballSize, BLACK);  // x, y, radius
    display.drawBitmap(tenX, tenY, tennis_logo, 128, 64, BLACK);
    xPos += 3;
    ballSize += 3.8;
  }

  for (int i = 0; i < 22; i++) {
    display.drawBitmap(2, 80 - (Y + changeY), amomii_logo, 128, 64, WHITE);
    display.display();
    display.drawBitmap(2, 80 - (Y + changeY), amomii_logo, 128, 64, BLACK);

    changeY = changeY + yVelocity;
    yVelocity -= 2;
    Y += 16;
  }
  display.drawBitmap(0, 0, nt_logo, 128, 64, WHITE);
  display.display();
  display.clearDisplay();
}

void playSuccessJingle() {
  // Play a quick success jingle
  tone(buzzer, 880, 100);  // A4
  delay(100);
  tone(buzzer, 988, 100);  // B4
  delay(100);
  tone(buzzer, 1047, 100);  // C5
  delay(100);
  noTone(buzzer);
}

void loop() {
  // Prompt user to press Key1 to start the test
  Serial.println(F("Press Key1 to start the test. You should see an image on the amomii Blink screen."));
  while (digitalRead(button[0]) == HIGH) {
    // Wait for the user to press Key1
  }
  playSuccessJingle();  // Play success jingle
  delay(200);           // Debounce delay
  while (digitalRead(button[0]) == LOW) {
    // Wait for the user to release Key1
  }
  tennisIntro();

  // Ask if an image showed on the screen
  Serial.println(F("Did you see an image on the screen? Press Key1 to continue."));
  while (digitalRead(button[0]) == HIGH) {
    // Wait for the user to press Key1
  }
  delay(200);  // Debounce delay
  while (digitalRead(button[0]) == LOW) {
    // Wait for the user to release Key1
  }
  playSuccessJingle();  // Play success jingle

  // Test LEDs and buttons
  Serial.println(F("Press each key (Key1 to Key8) sequentially to test the LEDs and buttons."));
  int buttonIndex = 0;
  int pattern = B10000000;
  for (int i = 0; i < 8; i++) {
    while (digitalRead(button[buttonIndex]) == HIGH)
      ;
    updateLEDs(pattern, pattern, pattern, pattern);
    buttonIndex++;
    pattern = pattern >> 1;
    tone(buzzer, 880, 20);
    delay(200);  // To avoid bouncing effects
  }

  // Confirm that buttons and LEDs are working properly
  Serial.println(F("Did each LED light up sequentially, and did you hear beeps? Press Key1 to continue."));
  while (digitalRead(button[0]) == HIGH) {
    // Wait for the user to press Key1
  }
  delay(200);  // Debounce delay
  while (digitalRead(button[0]) == LOW) {
    // Wait for the user to release Key1
  }
  playSuccessJingle();  // Play success jingle

  updateLEDs(0, 0, 0, 0);

  Serial.println(F("Now, press Key1 to test the amomii Glow strip with a rainbow chase effect."));
  while (digitalRead(button[0]) == HIGH) {
    // Wait for the user to press Key1
  }

  playSuccessJingle();  // Play success jingle
  Serial.println(F("Test Complete!"));
  Serial.println(F("If everything is working as expected, upload the Neon Tennis sketch and get your game on!"));


  // Rainbow chase effect
  while (true) {
    for (int i = 0; i < NUM_BUTTONS; i++) {
      for (int j = 0; j < NUM_LEDS; j++) {
        leds[j] = CHSV((j * 255 / NUM_LEDS), 255, 255);
      }
      FastLED.show();
      delay(50);
    }
  }

  Serial.println(F("The amomii Glow strip should be glowing with a rainbow chase effect. Test complete."));
}
